@@ -1,30 +1,45 @@
 #!/usr/bin/env python
 # coding: utf-8
 
-# Copyright (c) Saga Inc.
-# Distributed under the terms of the GPL License.
+# Copyright (c) Mito.
+# Distributed under the terms of the Modified BSD License.
+
 """
-Exports an error box that can be used to pass the errors we need. 
+Exports an error box that can be used to pass the errors
+we need. 
 
 Further exports helper functions for creating errors
 of different flavors. We wrap these functions in the same
-MitoError box to avoid creating too many new classes!
+EditError box to avoid creating too many new classes!
 
 See more about why we use errors here: 
 https://stackoverflow.com/questions/16138232/is-it-a-good-practice-to-use-try-except-else-in-python
 """
-import re
 import traceback
-from typing import Any, Collection, Optional, Set, List
+from typing import Set, List
 
-from mitosheet.types import ColumnHeader
+class CreationError(Exception):
+    """
+    An error that occurs during a mito.sheet call, on creation
+    of the mito widget. 
+    """
+    def __init__(self, type_, header, to_fix):
+        """
+        Creates a creation error. 
 
+        type_: the error id string.
+        header: the header of the error message to be displayed.
+        to_fix: instructions on how to handle or fix the error.
+        """
+        self.type_ = type_ # we have an _ to avoid overwriting the build in type
+        self.header = header
+        self.to_fix = to_fix
 
-class MitoError(Exception):
+class EditError(Exception):
     """
     An error that occurs during the processing of an editing event.
     """
-    def __init__(self, type_: str, header: str, to_fix: str, error_modal: bool=True):
+    def __init__(self, type_, header, to_fix):
         """
         Creates a creation error. 
 
@@ -35,10 +50,8 @@ class MitoError(Exception):
         self.type_ = type_ # we have an _ to avoid overwriting the build in type
         self.header = header
         self.to_fix = to_fix
-        self.error_modal = error_modal
-        self.traceback = get_recent_traceback() # record the most recent error when the error is created
 
-def make_no_sheet_error(sheet_indexes: Set[int]) -> MitoError:
+def make_no_sheet_error(sheet_indexes: Set[str]):
     """
     Helper function for creating a no_sheet_error.
 
@@ -46,53 +59,13 @@ def make_no_sheet_error(sheet_indexes: Set[int]) -> MitoError:
     """
     to_fix = f'There is an issue behind the scenes with that operation, we\'ll get it fixed as soon as possible!'
 
-    return MitoError(
+    return EditError(
         'no_sheet_error', 
         'No Sheet Exists',
         to_fix
     )
 
-def make_incompatible_merge_headers_error(error_modal: bool=True) -> MitoError:
-    """
-    Helper function for creating a incompatible_merge_headers_error
-
-    Occurs when a user merges on two dataframes, where one has a multi-index header
-    and one does not.
-    """
-    to_fix = f'Invalid merge, as a dataframe has multi-index headers. To fix, edit your pivot table and flatten the headers.'
-
-    return MitoError(
-        'incompatible_merge_key_error', 
-        'Incompatible Merge Key Types',
-        to_fix,
-        error_modal=error_modal
-    )
-
-def make_incompatible_merge_key_error(merge_key_one: Optional[ColumnHeader]=None, merge_key_one_dtype: Optional[str]=None, merge_key_two: Optional[ColumnHeader]=None, merge_key_two_dtype: Optional[str]=None, error_modal: bool=True) -> MitoError:
-    """
-    Helper function for creating a incompatible_merge_key_error
-
-    Occurs when a user merges on two keys that are different types
-    """
-    from mitosheet.column_headers import get_column_header_display
-
-    if merge_key_one is not None and merge_key_two is not None:
-        return MitoError(
-            'incompatible_merge_key_error', 
-            'Incompatible Merge Key Types',
-            f'{get_column_header_display(merge_key_one)} ({merge_key_one_dtype}) and {get_column_header_display(merge_key_two)} ({merge_key_two_dtype}) have different types. Either pick new keys or cast their types.',
-            error_modal=error_modal
-        )
-    
-    to_fix = f'You cannot merge using keys of different types. Either pick new keys or cast their types.'
-    return MitoError(
-        'incompatible_merge_key_error', 
-        'Incompatible Merge Key Types',
-        to_fix,
-        error_modal=error_modal
-    )
-
-def make_no_column_error(column_headers: Collection[ColumnHeader], error_modal: bool=True) -> MitoError:
+def make_no_column_error(column_headers: Set[str]):
     """
     Helper function for creating a no_column_error.
 
@@ -101,29 +74,28 @@ def make_no_column_error(column_headers: Collection[ColumnHeader], error_modal:
     if len(column_headers) == 1:
         to_fix = f'Sorry, there is no column with the name {next(iter(column_headers))}. Did you type it correctly?'
     else:
-        to_fix = f'Sorry, there are no column with the names {", ".join(map(str, column_headers))}. Did you type them correctly?'
+        to_fix = f'Sorry, there are no column with the names {", ".join(column_headers)}. Did you type them correctly?'
 
-    return MitoError(
+    return EditError(
         'no_column_error', 
         'No Column Exists',
-        to_fix,
-        error_modal=error_modal
+        to_fix
     )
 
-def make_column_exists_error(column_header: ColumnHeader) -> MitoError:
+def make_column_exists_error(column_header: str):
     """
     Helper function for creating a column_exists_error.
 
     Occurs when:
     -  the user adds a column that already exists in the dataframe.
     """
-    return MitoError(
+    return EditError(
         'column_exists_error', 
         'Column Already Exists',
         f'Sorry, a column already exists with the name {column_header}. Try picking a different name!'
     )
 
-def make_invalid_formula_error(formula: str, to_fix: str=None, error_modal: bool=True) -> MitoError:
+def make_invalid_formula_error(formula: str, to_fix=None):
     """
     Helper function for creating a invalid_formula_error.
 
@@ -133,91 +105,54 @@ def make_invalid_formula_error(formula: str, to_fix: str=None, error_modal: bool
     if to_fix is None:
         to_fix = f'Sorry, the formula \'{formula}\' is invalid. Did you type it correctly?'
 
-    return MitoError(
+    return EditError(
         'invalid_formula_error',
         'Invalid Formula',
-        to_fix, 
-        error_modal=error_modal
+        to_fix
     )
 
-def make_invalid_formula_after_update_error(error_modal: bool=False) -> MitoError:
-    """
-    Helper function for creating a invalid_formula_after_update_error.
-
-    Occurs when a user edits a user tries to resubmit a formula that used to be
-    valid and is no longer valid - which must mean they changed something they 
-    rely on.
-    """
-  
-    return MitoError(
-        'invalid_formula_after_update_error', 
-        'Formula is Now Invalid',
-        'Renaming or deleting other columns in this dataset has made this formula invalid. Please update this formula.',
-        error_modal=error_modal
-    )
-
-
-def make_cast_value_to_type_error(value: str, column_dtype: str, error_modal: bool=True) -> MitoError:
-    """
-    Helper function for creating a cast_value_to_type_error.
-
-    Occurs when the user edits a cell to a specific value that cannot be cast to the 
-    type of the column it is in.
-
-    """
-    return MitoError(
-        'cast_value_to_type_error',
-        'Invalid Value for ' + column_dtype + 'Series',
-        'The value ' + value + ' could not be cast to the type ' + column_dtype + '. Please enter a different value.' , 
-        error_modal=error_modal
-    )
-
-def make_circular_reference_error(error_modal: bool=True) -> MitoError:
+def make_circular_reference_error():
     """
     Helper function for creating a circular_reference_error.
 
     Occurs when:
     -  an edit introduces a circular reference.
     """
-    return MitoError(
+    return EditError(
         'circular_reference_error',
         'Circular Reference',
-        f'Sorry, circular references are not supported currently.',
-        error_modal=error_modal
+        f'Sorry, circular references are not supported currently.'
     )
 
-def make_invalid_column_headers_error(column_headers: List[ColumnHeader]) -> MitoError:
+def make_wrong_column_metatype_error(column_header: str):
+    """
+    Helper function for creating a wrong_column_metatype_error.
+
+    Occurs when:
+    -  a user edits a column to a formula when it is not a formula type column.
+    """
+    return EditError(
+        'wrong_column_metatype_error',
+        'Wrong Column Type',
+        f'Sorry, the column {column_header} is a data column. We don\'t currently support changing it to a formula.'
+    )
+
+def make_invalid_column_headers_error(column_headers: List[str]):
     """
     Helper function for creating a invalid_column_headers_error.
 
     Occurs when:
     -  a user creates (or renames) a column(s) that has an invalid name.
     """
-    to_fix = f'All headers in the dataframe must contain at least one letter and no symbols other than numbers and "_". Invalid headers: {", ".join(map(str, column_headers))}'
+    to_fix = f'All headers in the dataframe must contain at least one letter and no symbols other than numbers and "_". Invalid headers: {", ".join(column_headers)}'
 
-    return MitoError(
+    return EditError(
         'invalid_column_header_error',
         'Invalid Column Header',
         to_fix
     )
 
-def make_function_error(function_name: str, error_modal: bool=True) -> MitoError:
-    """
-    Helper function for creating a function_error.
-
-    Occurs when there is an error inside a specific sheet function that
-    is running
-    """
-    # TODO: this to_fix message is _just terrible_
-    return MitoError(
-        'function_error',
-        'Error in ' + function_name,
-        f'Sorry, an unknown error occured in the ' + function_name + ' function. See the documentation for help.',
-        error_modal=error_modal
-    )
-
-
-def make_execution_error(error_modal: bool=True) -> MitoError:
+def make_execution_error():
     """
     Helper function for creating a execution_error.
 
@@ -225,14 +160,13 @@ def make_execution_error(error_modal: bool=True) -> MitoError:
     -  running an existing analysis on the parameters passed to the mito sheet throw an error.
     """
     # TODO: this to_fix message is _just terrible_
-    return MitoError(
+    return EditError(
         'execution_error',
         'Execution Error',
-        f'Sorry, there was an error during executing this code.',
-        error_modal=error_modal
+        f'Sorry, there was an error during executing this code.'
     )
 
-def make_function_execution_error(function: str) -> MitoError:
+def make_function_execution_error(function):
     """
     Helper function for creating a function_execution_error.
 
@@ -240,13 +174,13 @@ def make_function_execution_error(function: str) -> MitoError:
     -  An error occurs inside of a mito sheet function
     """
     # TODO: this to_fix message is _just terrible_
-    return MitoError(
+    return EditError(
         'execution_error',
         f'Error Executing {function}',
         f'Sorry, there was an error in the {function}. Please check the documentation to make sure you called it correctly.'
     )
 
-def make_unsupported_function_error(functions: Set[str], error_modal: bool=True) -> MitoError:
+def make_unsupported_function_error(functions: Set[str]):
     """
     Helper function for creating a unsupported_function_error.
 
@@ -259,50 +193,50 @@ def make_unsupported_function_error(functions: Set[str], error_modal: bool=True)
     else:
         to_fix = f'Sorry, mito does not currently support the functions {", ".join(functions)}. To request it, shoot us an email!'
 
-    return MitoError(
+    return EditError(
         'unsupported_function_error',
         'Unsupported Function',
-        to_fix,
-        error_modal=error_modal
+        to_fix
     )
 
-
-def make_invalid_column_delete_error(column_headers: Collection[ColumnHeader], dependents: Optional[Collection[ColumnHeader]]=None) -> MitoError:
+def make_invalid_column_delete_error(column_header, dependents):
     """
     Helper function for creating a invalid_column_delete_error.
 
     Occurs when:
-    - the user deletes a column that is referenced by other columns
-    - the user deletes a column that does not exist
+    -  the user deletes a column that is referenced by other columns
     """
-    if dependents is not None:
-        header = 'Column Has Dependents'
-        to_fix = f'{(", ").join(map(str, column_headers))} cannot be deleted, as {"they are" if len(column_headers) > 1 else "it is"} referenced in {(", ".join(map(str, dependents)))}. Please remove these references before deleting.'
-    else:
-        header = 'Column Does not exist'
-        to_fix = f'{(", ").join(map(str, column_headers))} cannot be deleted, as not all of these columns exist.'
-   
+    # We make sure it's a list, for easy accessing!
+    dependents = list(dependents)
 
-    return MitoError(
+    # We make the error message read nicely!
+    if len(dependents) == 1:
+        dependent_string = f'column {dependents[0]}'
+    elif len(dependents) == 2:
+        dependent_string = f'columns {dependents[0]} and {dependents[1]}'
+    else:
+        dependent_string = f'columns {", ".join(dependents[: -1])}, and {dependents[-1]}'
+
+    return EditError(
         'invalid_column_delete_error',
-        header,
-        to_fix
+        'Column Has Dependents',
+        f'{column_header} cannot be deleted, as it is referenced in {dependent_string}. Please remove these references before deleting.'
     )
 
-def make_invalid_arguments_error(function: str) -> MitoError:
+def make_invalid_arguments_error(function):
     """
     Helper function for creating a invalid_arguments_error.
 
     Occurs when:
     -  the user calls a sheet function with incorrect arguments
     """
-    return MitoError(
+    return EditError(
         'invalid_arguments_error',
         f'Invalid Arguments to {function}',
         f'It looks like you passed the wrong arguments to {function}. Try checking out the documentation to see correct usage!'
     )
 
-def make_invalid_aggregation_error() -> MitoError:
+def make_invalid_aggregation_error():
     """
     Helper function for creating a invalid_aggregation_error.
 
@@ -310,13 +244,13 @@ def make_invalid_aggregation_error() -> MitoError:
     -  the user tries to .agg with a function on a column that has the wrong type (e.g. 'mean' on a
         column of strings).
     """
-    return MitoError(
+    return EditError(
         'invalid_aggregation_error',
         f'Invalid Aggregation Function',
         f'Sorry, you tried to aggregate with a function that could not be applied to that type of column. Please try again!'
     )
 
-def make_invalid_filter_error(filter_value: Any, correct_type: str) -> MitoError:
+def make_invalid_filter_error(filter_value, correct_type):
     """
     Helper function for creating a invalid_filter_error.
 
@@ -332,13 +266,13 @@ def make_invalid_filter_error(filter_value: Any, correct_type: str) -> MitoError
     else:
         correct_format = 'that is valid.'
 
-    return MitoError(
+    return EditError(
         'invalid_filter_error',
         f'Invalid Filter',
         f'Sorry, the value {filter_value} is not a valid value for that {correct_type} filter. Please enter a value {correct_format}!'
     )
 
-def make_invalid_sort_error(column_header: ColumnHeader) -> MitoError:
+def make_invalid_sort_error(column_header):
     """
     Helper function for creating a invalid_sort_error.
 
@@ -347,13 +281,13 @@ def make_invalid_sort_error(column_header: ColumnHeader) -> MitoError:
     (e.g. a column with strings and floats).
     """
 
-    return MitoError(
+    return EditError(
         'invalid_sort_error',
         f'Invalid Sort',
         f'Sorry, the column {column_header} has mixed data types. Please make sure the column has one datatype before trying to sort.'
     )
 
-def make_df_exists_error(df_name: str) -> MitoError:
+def make_df_exists_error(df_name):
     """
     Helper function for creating a df_exists_error.
 
@@ -362,13 +296,13 @@ def make_df_exists_error(df_name: str) -> MitoError:
     dataframe in the sheet.
     """
 
-    return MitoError(
+    return EditError(
         'df_exists_error',
         f'Dataframe Exists',
         f'Sorry, the dataframe {df_name} already exists. Please pick a different name.'
     )
 
-def make_invalid_column_type_change_error(column_header: ColumnHeader, old_dtype: str, new_dtype: str) -> MitoError:
+def make_invalid_column_type_change_error(column_header, old_column_type, new_column_type):
     """
     Helper function for creating a invalid_column_type_change_error.
 
@@ -377,13 +311,13 @@ def make_invalid_column_type_change_error(column_header: ColumnHeader, old_dtype
       a number to a datetime, which is impossible).
     """
 
-    return MitoError(
+    return EditError(
         'invalid_column_type_change_error',
         f'Invalid Type Change',
-        f'Sorry, the column {column_header} has a type {old_dtype}, which cannot be changed to the type {new_dtype}.'
+        f'Sorry, the column {column_header} has a type {old_column_type}, which cannot be changed to the type {new_column_type}.'
     )
 
-def make_invalid_pivot_error() -> MitoError:
+def make_invalid_pivot_error():
     """
     Helper function for creating a invalid_pivot_error.
 
@@ -391,59 +325,39 @@ def make_invalid_pivot_error() -> MitoError:
     -  the user runs a pivot that is invalid in some way.
     """
     # TODO: this to_fix message is _just terrible_
-    return MitoError(
+    return EditError(
         'invalid_pivot_error',
         'Pivot Error',
         f'Sorry, there was an error computing your pivot. Please try a different pivot!'
     )
 
 
-def make_is_directory_error(file_name: str) -> MitoError:
+def make_is_directory_error(file_name):
     """
     Helper function for creating directory_error
 
     Occurs when:
     - the user tries to import a directory in a simple import
     """
-    return MitoError(
+    return EditError(
         'directory_error',
         f'Change Path to a File',
         f'Sorry, the path {file_name} is a directory. Please pass in a path to a file to import it into Mito.'
     )
 
 
-def make_no_analysis_error(analysis_id: str, error_modal: bool=True) -> MitoError:
+def make_duplicated_pivot_column_error(column_header):
     """
-    Helper function for creating a no_analysis_error.
+    Helper function for creating duplicated_pivot_column_error
 
-    Occurs when a user tries to replay an analysis that they don't have access to in 
-    their .mito folder
+    Occurs when:
+    - the user adds a column to both the row and the value section of the pivot table
+        which is not covered by a special case
     """
-    to_fix = f'When you call mitosheet.sheet() and there is a code cell below that contains an analysis ID, \
-    Mito tries to replay that analysis. However, you do not have access to the analysis {analysis_id} so Mito is unable to replay it. \
-    Add a new code cell below the mitosheet.sheet() call to start a new analysis.' 
-    
-    return MitoError(
-        'no_analysis_error', 
-        "You don't have access to the analysis",
-        to_fix,
-        error_modal=error_modal
-    )
-
-def make_invalid_promote_row_to_header(error_modal: bool=True) -> MitoError:
-    """
-    Helper function for creating a invalid_promote_row_to_header.
-
-    Occurs when a user tries to promote a row to header that has duplicated
-    values within it.
-    """
-    to_fix = f'This row has duplicated values in it. As such, making this row a header would make column references ambigious, so we cannot make it a header row.' 
-    
-    return MitoError(
-        'no_analysis_error', 
-        "Cannot create duplicate headers",
-        to_fix,
-        error_modal=error_modal
+    return EditError(
+        'duplicated_column_in_pivot_error',
+        f'Invalid Pivot Configuration',
+        f'Sorry, the column {column_header} is used twice. Please remove it from one section.'
     )
 
 ARG_FULL_NAME = {
@@ -453,7 +367,7 @@ ARG_FULL_NAME = {
     'bool': 'boolean'
 }
 
-def make_operator_type_error(operator: str, arg_one_type: str, arg_two_type: str) -> MitoError:
+def make_operator_type_error(operator, arg_one_type, arg_two_type):
     """
     Helper function for creating a operator_type_error.
 
@@ -474,22 +388,13 @@ def make_operator_type_error(operator: str, arg_one_type: str, arg_two_type: str
         # TODO: once we have a step that changes the column type, we should update this error to be more helpful
         to_fix = f'Sorry, you tried to use {operator} on a {type_one_full_name} and a {type_two_full_name}. Please change the type of one of these columns to be compatible.'
 
-    return MitoError(
+    return EditError(
         'operator_type_error',
         f'Error with {operator}',
         to_fix
     )
 
-def get_recent_traceback() -> str:
-    """
-    Helper function that returns the most recent traceback, with the file paths
-    stripped to remove all but the mitosheet file names for ease in debugging.
-    
-    Inspired by https://stackoverflow.com/questions/25272368/hide-file-from-traceback
-    """
-    return re.sub(r'File ".*[\\/]([^\\/]+.py)"', r'File "\1"', traceback.format_exc())
-
-def get_recent_traceback_as_list() -> List[str]:
+def get_recent_traceback_as_list():
     """
     Helper function for getting the traceback of the most recent error.
     
@@ -499,5 +404,4 @@ def get_recent_traceback_as_list() -> List[str]:
     We get this as a list of strings, as mixpanel truncates strings at 255 chars
     and thus we avoid things getting chopped this way.
     """
-    # NOTE: We ignore empty lines, as they add no information
-    return [line for line in get_recent_traceback().split('\n') if line != '']
+    return traceback.format_exc().split('\n')
